=== modified file '.rsyncexclude'
--- .rsyncexclude	2005-05-05 06:38:18 +0000
+++ .rsyncexclude	2005-10-10 01:58:41 +0000
@@ -15,3 +15,5 @@
 /tmp
 # do want this after all
 + CHANGELOG
+/build
+test*.tmp

=== modified file 'Makefile'
--- Makefile	2005-10-07 01:11:04 +0000
+++ Makefile	2005-10-09 23:12:35 +0000
@@ -10,6 +10,7 @@
 clean: executables
 	./setup.py clean
 	find . -name "*.pyc" | xargs rm
+	rm -rf test????.tmp
 
 .PHONY: all
 

=== modified file 'NEWS'
--- NEWS	2005-10-06 13:08:14 +0000
+++ NEWS	2005-10-10 08:29:42 +0000
@@ -1,5 +1,11 @@
 bzr 0.1 2005-10-06
 
+  NOTES:
+
+    * 'bzr branch' over http initially gives a very high estimate
+      of completion time but it should fall as the first few 
+      revisions are pulled in.  branch is still slow on 
+      high-latency connections.
 
   BUG FIXES:
   
@@ -9,6 +15,14 @@
     * Locking is now done with fcntl.lockf which works with NFS
       file systems. Contributed by Harald Meland.
 
+    * When a merge encounters a file that has been deleted on
+      one side and modified on the other, the old contents are
+      written out to foo.BASE and foo.SIDE, where SIDE is this
+      or OTHER. Contributed by Aaron Bentley.
+
+    * Export was choosing incorrect file paths for the content of
+      the tarball, this has been fixed by Aaron Bentley.
+
     * Commit will no longer commit without a log message, an 
       error is returned instead. Contributed by Jelmer Vernooij.
 
@@ -16,6 +30,20 @@
       parent directories that are added but not listed will be 
       automatically included. Suggested by Michael Ellerman.
 
+    * bzr commit and upgrade did not correctly record new revisions
+      for files with only a change to their executable status.
+      bzr will correct this when it encounters it. Fixed by
+      Robert Collins
+
+    * HTTP tests now force off the use of http_proxy for the duration.
+      Contributed by Gustavo Niemeyer.
+
+    * Fix problems in merging weave-based branches that have 
+      different partial views of history.
+
+    * Symlink support: working with symlinks when not in the root of a 
+      bzr tree was broken, patch from Scott James Remnant.
+
 
   IMPROVEMENTS:
 
@@ -41,6 +69,19 @@
       merge conflicts respectively. A conflicted tree cannot be committed
       in. Contributed by Aaron Bentley.
 
+    * 'rm' is now an alias for 'remove'.
+
+    * Stores now split out their content in a single byte prefixed hash,
+      dropping the density of files per directory by 256. Contributed by
+      Gustavo Niemeyer.
+
+    * 'bzr diff -r branch:URL' will now perform a diff between two branches.
+      Contributed by Robert Collins.
+
+    * 'bzr log' with the default formatter will show merged revisions,
+      indented to the right. Initial implementation contributed by Gustavo
+      Niemeyer, made incremental by Robert Collins.
+
 
   INTERNALS:
 
@@ -50,6 +91,10 @@
     * InventoryEntry is now an abstract base class, use one of the
       concrete InventoryDirectory etc classes instead.
 
+    * Branch raises an UnsupportedFormatError when it detects a 
+      bzr branch it cannot understand. This allows for precise
+      handling of such circumstances.
+
 
   TESTING:
 

=== modified file 'bzr' (properties changed: -x to +x)
=== modified file 'bzrlib/__init__.py'
--- bzrlib/__init__.py	2005-10-06 12:34:22 +0000
+++ bzrlib/__init__.py	2005-10-10 04:30:52 +0000
@@ -39,7 +39,7 @@
 del locale
 
 __copyright__ = "Copyright 2005 Canonical Development Ltd."
-__version__ = '0.1rc1'
+__version__ = '0.1rc2'
 
 
 def get_bzr_revision():

=== modified file 'bzrlib/add.py'
--- bzrlib/add.py	2005-10-03 02:10:02 +0000
+++ bzrlib/add.py	2005-10-08 00:49:13 +0000
@@ -14,12 +14,14 @@
 # along with this program; if not, write to the Free Software
 # Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 
+from os.path import dirname
+
+import bzrlib.errors as errors
 from bzrlib.inventory import InventoryEntry
 from bzrlib.trace import mutter, note, warning
 from bzrlib.errors import NotBranchError
 from bzrlib.branch import Branch
 from bzrlib.osutils import quotefn
-from os.path import dirname
 
 def glob_expand_for_win32(file_list):
     import glob
@@ -116,6 +118,8 @@
                 sub_tree = True
             except NotBranchError:
                 sub_tree = False
+            except errors.UnsupportedFormatError:
+                sub_tree = True
         else:
             sub_tree = False
 

=== modified file 'bzrlib/branch.py'
--- bzrlib/branch.py	2005-10-07 06:13:14 +0000
+++ bzrlib/branch.py	2005-10-10 21:52:15 +0000
@@ -51,6 +51,7 @@
 
 BZR_BRANCH_FORMAT_4 = "Bazaar-NG branch, format 0.0.4\n"
 BZR_BRANCH_FORMAT_5 = "Bazaar-NG branch, format 5\n"
+BZR_BRANCH_FORMAT_6 = "Bazaar-NG branch, format 6\n"
 ## TODO: Maybe include checks for common corruption of newlines, etc?
 
 
@@ -240,24 +241,26 @@
             self._make_control()
         self._check_format(relax_version_check)
 
-        def get_store(name, compressed=True):
+        def get_store(name, compressed=True, prefixed=False):
             # FIXME: This approach of assuming stores are all entirely compressed
             # or entirely uncompressed is tidy, but breaks upgrade from 
             # some existing branches where there's a mixture; we probably 
             # still want the option to look for both.
             relpath = self._rel_controlfilename(name)
             if compressed:
-                store = CompressedTextStore(self._transport.clone(relpath))
+                store = CompressedTextStore(self._transport.clone(relpath),
+                                            prefixed=prefixed)
             else:
-                store = TextStore(self._transport.clone(relpath))
+                store = TextStore(self._transport.clone(relpath),
+                                  prefixed=prefixed)
             #if self._transport.should_cache():
             #    cache_path = os.path.join(self.cache_root, name)
             #    os.mkdir(cache_path)
             #    store = bzrlib.store.CachedStore(store, cache_path)
             return store
-        def get_weave(name):
+        def get_weave(name, prefixed=False):
             relpath = self._rel_controlfilename(name)
-            ws = WeaveStore(self._transport.clone(relpath))
+            ws = WeaveStore(self._transport.clone(relpath), prefixed=prefixed)
             if self._transport.should_cache():
                 ws.enable_cache = True
             return ws
@@ -270,6 +273,11 @@
             self.control_weaves = get_weave([])
             self.weave_store = get_weave('weaves')
             self.revision_store = get_store('revision-store', compressed=False)
+        elif self._branch_format == 6:
+            self.control_weaves = get_weave([])
+            self.weave_store = get_weave('weaves', prefixed=True)
+            self.revision_store = get_store('revision-store', compressed=False,
+                                            prefixed=True)
         self._transaction = None
 
     def __str__(self):
@@ -335,6 +343,7 @@
         self._transaction = new_transaction
 
     def lock_write(self):
+        mutter("lock write: %s (%s)", self, self._lock_count)
         # TODO: Upgrade locking to support using a Transport,
         # and potentially a remote locking protocol
         if self._lock_mode:
@@ -351,6 +360,7 @@
 
 
     def lock_read(self):
+        mutter("lock read: %s (%s)", self, self._lock_count)
         if self._lock_mode:
             assert self._lock_mode in ('r', 'w'), \
                    "invalid lock mode %r" % self._lock_mode
@@ -363,6 +373,7 @@
             self._set_transaction(transactions.ReadOnlyTransaction())
                         
     def unlock(self):
+        mutter("unlock: %s (%s)", self, self._lock_count)
         if not self._lock_mode:
             raise LockError('branch %r is not locked' % (self))
 
@@ -471,7 +482,7 @@
         files = [('README', 
             "This is a Bazaar-NG control directory.\n"
             "Do not change any files in this directory.\n"),
-            ('branch-format', BZR_BRANCH_FORMAT_5),
+            ('branch-format', BZR_BRANCH_FORMAT_6),
             ('revision-history', ''),
             ('branch-name', ''),
             ('branch-lock', ''),
@@ -499,14 +510,17 @@
         except NoSuchFile:
             raise NotBranchError(self.base)
         mutter("got branch format %r", fmt)
-        if fmt == BZR_BRANCH_FORMAT_5:
+        if fmt == BZR_BRANCH_FORMAT_6:
+            self._branch_format = 6
+        elif fmt == BZR_BRANCH_FORMAT_5:
             self._branch_format = 5
         elif fmt == BZR_BRANCH_FORMAT_4:
             self._branch_format = 4
 
         if (not relax_version_check
-            and self._branch_format != 5):
-            raise BzrError('sorry, branch format %r not supported' % fmt,
+            and self._branch_format not in (5, 6)):
+            raise errors.UnsupportedFormatError(
+                           'sorry, branch format %r not supported' % fmt,
                            ['use a different bzr version',
                             'or remove the .bzr directory'
                             ' and "bzr init" again'])

=== modified file 'bzrlib/builtins.py'
--- bzrlib/builtins.py	2005-10-06 13:08:14 +0000
+++ bzrlib/builtins.py	2005-10-08 08:45:42 +0000
@@ -475,6 +475,7 @@
     """
     takes_args = ['file+']
     takes_options = ['verbose']
+    aliases = ['rm']
     
     def run(self, file_list, verbose=False):
         b = Branch.open_containing(file_list[0])

=== modified file 'bzrlib/changeset.py'
--- bzrlib/changeset.py	2005-10-03 19:50:36 +0000
+++ bzrlib/changeset.py	2005-10-10 03:42:29 +0000
@@ -45,50 +45,57 @@
     return newdict
 
        
-class ChangeUnixPermissions(object):
+class ChangeExecFlag(object):
     """This is two-way change, suitable for file modification, creation,
     deletion"""
-    def __init__(self, old_mode, new_mode):
-        self.old_mode = old_mode
-        self.new_mode = new_mode
+    def __init__(self, old_exec_flag, new_exec_flag):
+        self.old_exec_flag = old_exec_flag
+        self.new_exec_flag = new_exec_flag
 
     def apply(self, filename, conflict_handler, reverse=False):
         if not reverse:
-            from_mode = self.old_mode
-            to_mode = self.new_mode
+            from_exec_flag = self.old_exec_flag
+            to_exec_flag = self.new_exec_flag
         else:
-            from_mode = self.new_mode
-            to_mode = self.old_mode
+            from_exec_flag = self.new_exec_flag
+            to_exec_flag = self.old_exec_flag
         try:
-            current_mode = os.stat(filename).st_mode &0777
+            current_exec_flag = bool(os.stat(filename).st_mode & 0111)
         except OSError, e:
             if e.errno == errno.ENOENT:
-                if conflict_handler.missing_for_chmod(filename) == "skip":
+                if conflict_handler.missing_for_exec_flag(filename) == "skip":
                     return
                 else:
-                    current_mode = from_mode
+                    current_exec_flag = from_exec_flag
 
-        if from_mode is not None and current_mode != from_mode:
-            if conflict_handler.wrong_old_perms(filename, from_mode, 
-                                                current_mode) != "continue":
+        if from_exec_flag is not None and current_exec_flag != from_exec_flag:
+            if conflict_handler.wrong_old_exec_flag(filename,
+                        from_exec_flag, current_exec_flag) != "continue":
                 return
 
-        if to_mode is not None:
+        if to_exec_flag is not None:
+            current_mode = os.stat(filename).st_mode
+            if to_exec_flag:
+                umask = os.umask(0)
+                os.umask(umask)
+                to_mode = current_mode | (0100 & ~umask)
+                # Enable x-bit for others only if they can read it.
+                if current_mode & 0004:
+                    to_mode |= 0001 & ~umask
+                if current_mode & 0040:
+                    to_mode |= 0010 & ~umask
+            else:
+                to_mode = current_mode & ~0111
             try:
                 os.chmod(filename, to_mode)
             except IOError, e:
                 if e.errno == errno.ENOENT:
-                    conflict_handler.missing_for_chmod(filename)
+                    conflict_handler.missing_for_exec_flag(filename)
 
     def __eq__(self, other):
-        if not isinstance(other, ChangeUnixPermissions):
-            return False
-        elif self.old_mode != other.old_mode:
-            return False
-        elif self.new_mode != other.new_mode:
-            return False
-        else:
-            return True
+        return (isinstance(other, ChangeExecFlag) and
+                self.old_exec_flag == other.old_exec_flag and
+                self.new_exec_flag == other.new_exec_flag)
 
     def __ne__(self, other):
         return not (self == other)
@@ -916,31 +923,16 @@
         Exception.__init__(self, "Conflict applying changes to %s" % this_path)
         self.this_path = this_path
 
-class MergePermissionConflict(Exception):
-    def __init__(self, this_path, base_path, other_path):
-        this_perms = os.stat(this_path).st_mode & 0755
-        base_perms = os.stat(base_path).st_mode & 0755
-        other_perms = os.stat(other_path).st_mode & 0755
-        msg = """Conflicting permission for %s
-this: %o
-base: %o
-other: %o
-        """ % (this_path, this_perms, base_perms, other_perms)
-        self.this_path = this_path
-        self.base_path = base_path
-        self.other_path = other_path
-        Exception.__init__(self, msg)
-
 class WrongOldContents(Exception):
     def __init__(self, filename):
         msg = "Contents mismatch deleting %s" % filename
         self.filename = filename
         Exception.__init__(self, msg)
 
-class WrongOldPermissions(Exception):
-    def __init__(self, filename, old_perms, new_perms):
-        msg = "Permission missmatch on %s:\n" \
-        "Expected 0%o, got 0%o." % (filename, old_perms, new_perms)
+class WrongOldExecFlag(Exception):
+    def __init__(self, filename, old_exec_flag, new_exec_flag):
+        msg = "Executable flag missmatch on %s:\n" \
+        "Expected %s, got %s." % (filename, old_exec_flag, new_exec_flag)
         self.filename = filename
         Exception.__init__(self, msg)
 
@@ -964,7 +956,7 @@
         Exception.__init__(self, msg)
         self.filename = filename
 
-class MissingPermsFile(Exception):
+class MissingForSetExec(Exception):
     def __init__(self, filename):
         msg = "Attempt to change permissions on  %s, which does not exist" %\
             filename
@@ -1027,17 +1019,14 @@
         os.unlink(new_file)
         raise MergeConflict(this_path)
 
-    def permission_conflict(self, this_path, base_path, other_path):
-        raise MergePermissionConflict(this_path, base_path, other_path)
-
     def wrong_old_contents(self, filename, expected_contents):
         raise WrongOldContents(filename)
 
     def rem_contents_conflict(self, filename, this_contents, base_contents):
         raise RemoveContentsConflict(filename)
 
-    def wrong_old_perms(self, filename, old_perms, new_perms):
-        raise WrongOldPermissions(filename, old_perms, new_perms)
+    def wrong_old_exec_flag(self, filename, old_exec_flag, new_exec_flag):
+        raise WrongOldExecFlag(filename, old_exec_flag, new_exec_flag)
 
     def rmdir_non_empty(self, filename):
         raise DeletingNonEmptyDirectory(filename)
@@ -1048,8 +1037,8 @@
     def patch_target_missing(self, filename, contents):
         raise PatchTargetMissing(filename)
 
-    def missing_for_chmod(self, filename):
-        raise MissingPermsFile(filename)
+    def missing_for_exec_flag(self, filename):
+        raise MissingForExecFlag(filename)
 
     def missing_for_rm(self, filename, change):
         raise MissingForRm(filename)
@@ -1265,9 +1254,9 @@
         return new_meta
     elif new_meta is None:
         return old_meta
-    elif isinstance(old_meta, ChangeUnixPermissions) and \
-        isinstance(new_meta, ChangeUnixPermissions):
-        return ChangeUnixPermissions(old_meta.old_mode, new_meta.new_mode)
+    elif (isinstance(old_meta, ChangeExecFlag) and
+          isinstance(new_meta, ChangeExecFlag)):
+        return ChangeExecFlag(old_meta.old_exec_flag, new_meta.new_exec_flag)
     else:
         return ApplySequence(old_meta, new_meta)
 
@@ -1391,7 +1380,7 @@
         stat_a = self.lstat(full_path_a)
         stat_b = self.lstat(full_path_b)
 
-        cs_entry.metadata_change = self.make_mode_change(stat_a, stat_b)
+        cs_entry.metadata_change = self.make_exec_flag_change(stat_a, stat_b)
 
         if id in self.tree_a and id in self.tree_b:
             a_sha1 = self.tree_a.get_file_sha1(id)
@@ -1405,16 +1394,15 @@
                                                              stat_b)
         return cs_entry
 
-    def make_mode_change(self, stat_a, stat_b):
-        mode_a = None
+    def make_exec_flag_change(self, stat_a, stat_b):
+        exec_flag_a = exec_flag_b = None
         if stat_a is not None and not stat.S_ISLNK(stat_a.st_mode):
-            mode_a = stat_a.st_mode & 0777
-        mode_b = None
+            exec_flag_a = bool(stat_a.st_mode & 0111)
         if stat_b is not None and not stat.S_ISLNK(stat_b.st_mode):
-            mode_b = stat_b.st_mode & 0777
-        if mode_a == mode_b:
+            exec_flag_b = bool(stat_b.st_mode & 0111)
+        if exec_flag_a == exec_flag_b:
             return None
-        return ChangeUnixPermissions(mode_a, mode_b)
+        return ChangeExecFlag(exec_flag_a, exec_flag_b)
 
     def make_contents_change(self, full_path_a, stat_a, full_path_b, stat_b):
         if stat_a is None and stat_b is None:

=== modified file 'bzrlib/clone.py'
--- bzrlib/clone.py	2005-10-06 06:31:33 +0000
+++ bzrlib/clone.py	2005-10-10 03:35:19 +0000
@@ -75,24 +75,28 @@
     assert isinstance(to_location, basestring)
     if basis_branch is not None:
         note("basis_branch is not supported for fast weave copy yet.")
-    if not (branch_from.weave_store.listable()
-            and branch_from.revision_store.listable()):
-        return copy_branch_slower(branch_from, to_location, revision,
-                                  basis_branch)
-    history = _get_truncated_history(branch_from, revision)
-    if not bzrlib.osutils.lexists(to_location):
-        os.mkdir(to_location)
-    branch_to = Branch.initialize(to_location)
-    mutter("copy branch from %s to %s", branch_from, branch_to)
-    branch_to.set_root_id(branch_from.get_root_id())
-    branch_to.append_revision(*history)
-    _copy_control_weaves(branch_from, branch_to)
-    _copy_text_weaves(branch_from, branch_to)
-    _copy_revision_store(branch_from, branch_to)
-    build_working_dir(to_location)
-    branch_to.set_parent(branch_from.base)
-    mutter("copied")
-    return branch_to
+    branch_from.lock_read()
+    try:
+        if not (branch_from.weave_store.listable()
+                and branch_from.revision_store.listable()):
+            return copy_branch_slower(branch_from, to_location, revision,
+                                      basis_branch)
+        history = _get_truncated_history(branch_from, revision)
+        if not bzrlib.osutils.lexists(to_location):
+            os.mkdir(to_location)
+        branch_to = Branch.initialize(to_location)
+        mutter("copy branch from %s to %s", branch_from, branch_to)
+        branch_to.set_root_id(branch_from.get_root_id())
+        branch_to.append_revision(*history)
+        _copy_control_weaves(branch_from, branch_to)
+        _copy_text_weaves(branch_from, branch_to)
+        _copy_revision_store(branch_from, branch_to)
+        build_working_dir(to_location)
+        branch_to.set_parent(branch_from.base)
+        mutter("copied")
+        return branch_to
+    finally:
+        branch_from.unlock()
 
 
 def _get_truncated_history(branch_from, revision):

=== modified file 'bzrlib/commands.py'
--- bzrlib/commands.py	2005-10-06 13:08:14 +0000
+++ bzrlib/commands.py	2005-10-10 01:32:08 +0000
@@ -71,7 +71,7 @@
     """This handles a revision string -> revno.
 
     This always returns a list.  The list will have one element for
-    each revision.
+    each revision specifier supplied.
 
     >>> _parse_revision_str('234')
     [<RevisionSpec_int 234>]
@@ -85,7 +85,7 @@
     [<RevisionSpec_int 234>, <RevisionSpec None>]
     >>> _parse_revision_str('234..456..789') # Maybe this should be an error
     [<RevisionSpec_int 234>, <RevisionSpec_int 456>, <RevisionSpec_int 789>]
-    >>> _parse_revision_str('234....789') # Error?
+    >>> _parse_revision_str('234....789') #Error ?
     [<RevisionSpec_int 234>, <RevisionSpec None>, <RevisionSpec_int 789>]
     >>> _parse_revision_str('revid:test@other.com-234234')
     [<RevisionSpec_revid revid:test@other.com-234234>]
@@ -113,6 +113,8 @@
     Traceback (most recent call last):
       ...
     BzrError: No namespace registered for string: 'abc'
+    >>> _parse_revision_str('branch:../branch2')
+    [<RevisionSpec_branch branch:../branch2>]
     """
     import re
     old_format_re = re.compile('\d*:\d*')
@@ -127,11 +129,20 @@
             else:
                 revs.append(RevisionSpec(None))
     else:
+        next_prefix = None
         for x in revstr.split('..'):
             if not x:
                 revs.append(RevisionSpec(None))
+            elif x[-1] == ':':
+                # looks like a namespace:.. has happened
+                next_prefix = x + '..'
             else:
+                if next_prefix is not None:
+                    x = next_prefix + x
                 revs.append(RevisionSpec(x))
+                next_prefix = None
+        if next_prefix is not None:
+            revs.append(RevisionSpec(next_prefix))
     return revs
 
 

=== modified file 'bzrlib/diff.py'
--- bzrlib/diff.py	2005-10-03 05:16:47 +0000
+++ bzrlib/diff.py	2005-10-10 01:32:08 +0000
@@ -140,7 +140,7 @@
         oldtmpf.close()                 # and delete
         newtmpf.close()
 
-def show_diff(b, revision, specific_files, external_diff_options=None,
+def show_diff(b, from_spec, specific_files, external_diff_options=None,
               revision2=None, output=None):
     """Shortcut for showing the diff to the working tree.
 
@@ -148,7 +148,7 @@
         Branch.
 
     revision
-        None for each, or otherwise the old revision to compare against.
+        None for 'basis tree', or otherwise the old revision to compare against.
     
     The more general form is show_diff_trees(), where the caller
     supplies any two trees.
@@ -157,10 +157,10 @@
         import sys
         output = sys.stdout
 
-    if revision is None:
+    if from_spec is None:
         old_tree = b.basis_tree()
     else:
-        old_tree = b.revision_tree(revision.in_history(b).rev_id)
+        old_tree = b.revision_tree(from_spec.in_history(b).rev_id)
 
     if revision2 is None:
         new_tree = b.working_tree()

=== modified file 'bzrlib/errors.py'
--- bzrlib/errors.py	2005-10-07 06:13:14 +0000
+++ bzrlib/errors.py	2005-10-08 00:49:13 +0000
@@ -65,6 +65,12 @@
         return 'not a branch: %s' % self.args[0]
 
 
+class UnsupportedFormatError(BzrError):
+    """Specified path is a bzr branch that we cannot read."""
+    def __str__(self):
+        return 'unsupported branch format: %s' % self.args[0]
+
+
 class NotVersionedError(BzrError):
     """Specified object is not versioned."""
 
@@ -197,7 +203,7 @@
         BzrError.__init__(self, "Stores for branch %s are not listable" % br)
 
 
-from bzrlib.weave import WeaveError
+from bzrlib.weave import WeaveError, WeaveParentMismatch
 
 class TransportError(BzrError):
     """All errors thrown by Transport implementations should derive

=== modified file 'bzrlib/fetch.py'
--- bzrlib/fetch.py	2005-10-07 06:13:14 +0000
+++ bzrlib/fetch.py	2005-10-10 21:52:15 +0000
@@ -17,13 +17,14 @@
 import os
 from cStringIO import StringIO
 
-import bzrlib.errors
+import bzrlib
+import bzrlib.errors as errors
+from bzrlib.errors import InstallFailed, NoSuchRevision, WeaveError
 from bzrlib.trace import mutter, note, warning
 from bzrlib.branch import Branch
 from bzrlib.progress import ProgressBar
 from bzrlib.xml5 import serializer_v5
 from bzrlib.osutils import sha_string, split_lines
-from bzrlib.errors import InstallFailed, NoSuchRevision, WeaveError
 
 """Copying of history from one branch to another.
 
@@ -102,6 +103,14 @@
             self.pb = bzrlib.ui.ui_factory.progress_bar()
         else:
             self.pb = pb
+        self.from_branch.lock_read()
+        try:
+            self._fetch_revisions(last_revision)
+        finally:
+            self.from_branch.unlock()
+            self.pb.clear()
+
+    def _fetch_revisions(self, last_revision):
         try:
             self.last_revision = self._find_last_revision(last_revision)
         except NoSuchRevision, e:
@@ -115,7 +124,6 @@
         self._copy_revisions(revs_to_fetch)
         self.new_ancestry = revs_to_fetch
 
-
     def _find_last_revision(self, last_revision):
         """Find the limiting source revision.
 
@@ -218,11 +226,14 @@
         if file_id in self.copied_file_ids:
             mutter('file {%s} already copied', file_id)
             return
-        from_weave = self.from_weaves.get_weave(file_id, 
+        from_weave = self.from_weaves.get_weave(file_id,
             self.from_branch.get_transaction())
         to_weave = self.to_weaves.get_weave_or_empty(file_id,
             self.to_branch.get_transaction())
-        to_weave.join(from_weave)
+        try:
+            to_weave.join(from_weave)
+        except errors.WeaveParentMismatch:
+            to_weave.reweave(from_weave)
         self.to_weaves.put_weave(file_id, to_weave,
             self.to_branch.get_transaction())
         self.count_weaves += 1

=== modified file 'bzrlib/inventory.py'
--- bzrlib/inventory.py	2005-10-07 06:13:14 +0000
+++ bzrlib/inventory.py	2005-10-08 00:34:15 +0000
@@ -166,6 +166,14 @@
                 ie = inv[self.file_id]
                 assert ie.file_id == self.file_id
                 if ie.revision in heads:
+                    # fixup logic, there was a bug in revision updates.
+                    # with x bit support.
+                    try:
+                        if heads[ie.revision].executable != ie.executable:
+                            heads[ie.revision].executable = False
+                            ie.executable = False
+                    except AttributeError:
+                        pass
                     assert heads[ie.revision] == ie
                 else:
                     # may want to add it.
@@ -571,6 +579,8 @@
             # FIXME: 20050930 probe for the text size when getting sha1
             # in _read_tree_state
             self.text_size = previous_ie.text_size
+        if self.executable != previous_ie.executable:
+            compatible = False
         return compatible
 
 

=== modified file 'bzrlib/log.py'
--- bzrlib/log.py	2005-10-07 04:25:33 +0000
+++ bzrlib/log.py	2005-10-10 03:24:15 +0000
@@ -50,6 +50,7 @@
 """
 
 
+import bzrlib.errors as errors
 from bzrlib.tree import EmptyTree
 from bzrlib.delta import compare_trees
 from bzrlib.trace import mutter
@@ -206,6 +207,7 @@
     else:
         raise ValueError('invalid direction %r' % direction)
 
+    revision_history = branch.revision_history()
     for revno, rev_id in cut_revs:
         if verbose or specific_fileid:
             delta = _get_revision_delta(branch, revno)
@@ -225,7 +227,24 @@
                 continue
 
         lf.show(revno, rev, delta)
-
+        if revno == 1:
+            excludes = set()
+        else:
+            # revno is 1 based, so -2 to get back 1 less.
+            excludes = set(branch.get_ancestry(revision_history[revno - 2]))
+        pending = list(rev.parent_ids)
+        while pending:
+            rev_id = pending.pop()
+            if rev_id in excludes:
+                continue
+            # prevent showing merged revs twice if they multi-path.
+            excludes.add(rev_id)
+            try:
+                rev = branch.get_revision(rev_id)
+            except errors.NoSuchRevision:
+                continue
+            pending.extend(rev.parent_ids)
+            lf.show_merge(rev)
 
 
 def deltas_for_log_dummy(branch, which_revs):
@@ -321,6 +340,9 @@
     def show(self, revno, rev, delta):
         raise NotImplementedError('not implemented in abstract base')
 
+    def show_merge(self, rev):
+        pass
+
     
 class LongLogFormatter(LogFormatter):
     def show(self, revno, rev, delta):
@@ -353,6 +375,32 @@
         if delta != None:
             delta.show(to_file, self.show_ids)
 
+    def show_merge(self, rev):
+        from osutils import format_date
+
+        to_file = self.to_file
+
+        indent = '    '
+
+        print >>to_file,  indent+'-' * 60
+        print >>to_file,  indent+'merged:', rev.revision_id
+        if self.show_ids:
+            for parent_id in rev.parent_ids:
+                print >>to_file, indent+'parent:', parent_id
+            
+        print >>to_file,  indent+'committer:', rev.committer
+
+        date_str = format_date(rev.timestamp,
+                               rev.timezone or 0,
+                               self.show_timezone)
+        print >>to_file,  indent+'timestamp: %s' % date_str
+
+        print >>to_file,  indent+'message:'
+        if not rev.message:
+            print >>to_file,  indent+'  (no message)'
+        else:
+            for l in rev.message.split('\n'):
+                print >>to_file,  indent+'  ' + l
 
 
 class ShortLogFormatter(LogFormatter):

=== modified file 'bzrlib/merge.py'
--- bzrlib/merge.py	2005-10-04 04:32:32 +0000
+++ bzrlib/merge.py	2005-10-10 03:42:29 +0000
@@ -29,7 +29,7 @@
 from bzrlib.errors import BzrCommandError, UnrelatedBranches, NoCommonAncestor
 from bzrlib.errors import NoCommits
 from bzrlib.delta import compare_trees
-from bzrlib.trace import mutter, warning
+from bzrlib.trace import mutter, warning, note
 from bzrlib.fetch import greedy_fetch, fetch
 from bzrlib.revision import is_ancestor
 from bzrlib.osutils import rename
@@ -39,6 +39,8 @@
 # TODO: build_working_dir can be built on something simpler than merge()
 
 # FIXME: merge() parameters seem oriented towards the command line
+# NOTABUG: merge is a helper for commandline functions.  merge_inner is the
+#          the core functionality.
 
 # comments from abentley on irc: merge happens in two stages, each
 # of which generates a changeset object
@@ -140,6 +142,11 @@
     def rem_contents_conflict(self, filename, this_contents, base_contents):
         base_contents(filename+".BASE", self, False)
         this_contents(filename+".THIS", self, False)
+        return ReplaceContents(this_contents, None)
+
+    def rem_contents_conflict(self, filename, this_contents, base_contents):
+        base_contents(filename+".BASE", self, False)
+        this_contents(filename+".THIS", self, False)
         self.conflict("Other branch deleted locally modified file %s" %
                       filename)
         return ReplaceContents(this_contents, None)
@@ -185,7 +192,7 @@
 
     def finalize(self):
         if not self.ignore_zero:
-            print "%d conflicts encountered.\n" % self.conflicts
+            note("%d conflicts encountered.\n" % self.conflicts)
             
 def get_tree(treespec, temp_root, label, local_branch=None):
     location, revno = treespec
@@ -405,16 +412,6 @@
              source_file.interesting = source_file.id in interesting_ids
 
 
-def generate_cset_optimized(tree_a, tree_b, interesting_ids=None):
-    """Generate a changeset.  If interesting_ids is supplied, only changes
-    to those files will be shown.  Metadata changes are stripped.
-    """ 
-    cset =  generate_changeset(tree_a, tree_b, interesting_ids)
-    for entry in cset.entries.itervalues():
-        entry.metadata_change = None
-    return cset
-
-
 def merge_inner(this_branch, other_tree, base_tree, tempdir, 
                 ignore_zero=False, merge_type=ApplyMerge3, backup_files=False,
                 interesting_ids=None):
@@ -431,7 +428,7 @@
         return tree.tree.inventory
 
     inv_changes = merge_flex(this_tree, base_tree, other_tree,
-                             generate_cset_optimized, get_inventory,
+                             generate_changeset, get_inventory,
                              MergeConflictHandler(this_tree, base_tree,
                              other_tree, ignore_zero=ignore_zero),
                              merge_factory=merge_factory, 

=== modified file 'bzrlib/merge_core.py'
--- bzrlib/merge_core.py	2005-10-04 04:32:32 +0000
+++ bzrlib/merge_core.py	2005-10-10 03:42:29 +0000
@@ -260,18 +260,18 @@
     metadata = entry.metadata_change
     if metadata is None:
         return None
-    if isinstance(metadata, changeset.ChangeUnixPermissions):
-        if metadata.new_mode is None:
+    if isinstance(metadata, changeset.ChangeExecFlag):
+        if metadata.new_exec_flag is None:
             return None
-        elif metadata.old_mode is None:
+        elif metadata.old_exec_flag is None:
             return metadata
         else:
             base_path = base.readonly_path(entry.id)
             other_path = other.readonly_path(entry.id)    
-            return PermissionsMerge(base_path, other_path)
+            return ExecFlagMerge(base_path, other_path)
     
 
-class PermissionsMerge(object):
+class ExecFlagMerge(object):
     def __init__(self, base_path, other_path):
         self.base_path = base_path
         self.other_path = other_path
@@ -283,14 +283,26 @@
         else:
             base = self.other_path
             other = self.base_path
-        base_stat = os.stat(base).st_mode
-        other_stat = os.stat(other).st_mode
-        this_stat = os.stat(filename).st_mode
-        if base_stat &0777 == other_stat &0777:
-            return
-        elif this_stat &0777 == other_stat &0777:
-            return
-        elif this_stat &0777 == base_stat &0777:
-            os.chmod(filename, other_stat)
-        else:
-            conflict_handler.permission_conflict(filename, base, other)
+        base_mode = os.stat(base).st_mode
+        base_exec_flag = bool(base_mode & 0111)
+        other_mode = os.stat(other).st_mode
+        other_exec_flag = bool(other_mode & 0111)
+        this_mode = os.stat(filename).st_mode
+        this_exec_flag = bool(this_mode & 0111)
+        if (base_exec_flag != other_exec_flag and
+            this_exec_flag != other_exec_flag):
+            assert this_exec_flag == base_exec_flag
+            current_mode = os.stat(filename).st_mode
+            if other_exec_flag:
+                umask = os.umask(0)
+                os.umask(umask)
+                to_mode = current_mode | (0100 & ~umask)
+                # Enable x-bit for others only if they can read it.
+                if current_mode & 0004:
+                    to_mode |= 0001 & ~umask
+                if current_mode & 0040:
+                    to_mode |= 0010 & ~umask
+            else:
+                to_mode = current_mode & ~0111
+            os.chmod(filename, to_mode)
+

=== modified file 'bzrlib/revisionspec.py'
--- bzrlib/revisionspec.py	2005-09-29 20:34:25 +0000
+++ bzrlib/revisionspec.py	2005-10-10 01:32:08 +0000
@@ -151,6 +151,16 @@
         revs = branch.revision_history()
         return self._match_on_and_check(branch, revs)
 
+        # FIXME: in_history is somewhat broken,
+        # it will return non-history revisions in many
+        # circumstances. The expected facility is that
+        # in_history only returns revision-history revs,
+        # in_store returns any rev. RBC 20051010
+    # aliases for now, when we fix the core logic, then they
+    # will do what you expect.
+    in_store = in_history
+    in_branch = in_store
+        
     def __repr__(self):
         # this is mostly for helping with testing
         return '<%s %s%s>' % (self.__class__.__name__,
@@ -318,3 +328,27 @@
         return RevisionInfo(branch, revno, rev_id)
         
 SPEC_TYPES.append(RevisionSpec_ancestor)
+
+class RevisionSpec_branch(RevisionSpec):
+    """A branch: revision specifier.
+
+    This takes the path to a branch and returns its tip revision id.
+    """
+    prefix = 'branch:'
+
+    def _match_on(self, branch, revs):
+        from branch import Branch
+        from fetch import greedy_fetch
+        other_branch = Branch.open_containing(self.spec)
+        revision_b = other_branch.last_revision()
+        if revision_b is None:
+            raise NoCommits(other_branch)
+        # pull in the remote revisions so we can diff
+        greedy_fetch(branch, other_branch, revision=revision_b)
+        try:
+            revno = branch.revision_id_to_revno(revision_b)
+        except NoSuchRevision:
+            revno = None
+        return RevisionInfo(branch, revno, revision_b)
+        
+SPEC_TYPES.append(RevisionSpec_branch)

=== modified file 'bzrlib/selftest/HTTPTestUtil.py'
--- bzrlib/selftest/HTTPTestUtil.py	2005-09-29 20:34:25 +0000
+++ bzrlib/selftest/HTTPTestUtil.py	2005-10-09 22:59:39 +0000
@@ -137,9 +137,15 @@
         self._http_thread = threading.Thread(target=self._http_start)
         self._http_thread.setDaemon(True)
         self._http_thread.start()
+        self._http_proxy = os.environ.get("http_proxy")
+        if self._http_proxy is not None:
+            del os.environ["http_proxy"]
 
     def tearDown(self):
         self._http_running = False
         self._http_thread.join()
+        if self._http_proxy is not None:
+            import os
+            os.environ["http_proxy"] = self._http_proxy
         TestCaseInTempDir.tearDown(self)
 

=== modified file 'bzrlib/selftest/blackbox.py'
--- bzrlib/selftest/blackbox.py	2005-10-06 13:08:14 +0000
+++ bzrlib/selftest/blackbox.py	2005-10-10 01:32:08 +0000
@@ -31,6 +31,7 @@
 import os
 
 from bzrlib.branch import Branch
+from bzrlib.clone import copy_branch
 from bzrlib.errors import BzrCommandError
 from bzrlib.osutils import has_symlinks
 from bzrlib.selftest import TestCaseInTempDir, BzrTestBase
@@ -265,6 +266,25 @@
         output = self.runbzr('diff -r last:3..last:1', backtick=1)
         self.assert_('\n+baz' in output)
 
+    def test_diff_branches(self):
+        self.build_tree(['branch1/', 'branch1/file', 'branch2/'])
+        branch = Branch.initialize('branch1')
+        branch.add(['file'])
+        branch.commit('add file')
+        copy_branch(branch, 'branch2')
+        print >> open('branch2/file', 'w'), 'new content'
+        branch2 = Branch.open('branch2')
+        branch2.commit('update file')
+        # should open branch1 and diff against branch2, 
+        output = self.run_bzr_captured(['diff', '-r', 'branch:branch2', 'branch1'])
+        self.assertEquals(("=== modified file 'file'\n"
+                           "--- file\n"
+                           "+++ file\n"
+                           "@@ -1,1 +1,1 @@\n"
+                           "-new content\n"
+                           "+contents of branch1/file\n"
+                           "\n", ''), output)
+
     def test_branch(self):
         """Branch from one branch to another."""
         os.mkdir('a')
@@ -344,40 +364,6 @@
         assert os.path.exists('sub/a.txt.OTHER')
         assert os.path.exists('sub/a.txt.BASE')
 
-    def test_merge_with_missing_file(self):
-        """Merge handles missing file conflicts"""
-        os.mkdir('a')
-        os.chdir('a')
-        os.mkdir('sub')
-        print >> file('sub/a.txt', 'wb'), "hello"
-        print >> file('b.txt', 'wb'), "hello"
-        print >> file('sub/c.txt', 'wb'), "hello"
-        self.runbzr('init')
-        self.runbzr('add')
-        self.runbzr(('commit', '-m', 'added a'))
-        self.runbzr('branch . ../b')
-        print >> file('sub/a.txt', 'ab'), "there"
-        print >> file('b.txt', 'ab'), "there"
-        print >> file('sub/c.txt', 'ab'), "there"
-        self.runbzr(('commit', '-m', 'Added there'))
-        os.unlink('sub/a.txt')
-        os.unlink('sub/c.txt')
-        os.rmdir('sub')
-        os.unlink('b.txt')
-        self.runbzr(('commit', '-m', 'Removed a.txt'))
-        os.chdir('../b')
-        print >> file('sub/a.txt', 'ab'), "something"
-        print >> file('b.txt', 'ab'), "something"
-        print >> file('sub/c.txt', 'ab'), "something"
-        self.runbzr(('commit', '-m', 'Modified a.txt'))
-        self.runbzr('merge ../a/')
-        assert os.path.exists('sub/a.txt.THIS')
-        assert os.path.exists('sub/a.txt.BASE')
-        os.chdir('../a')
-        self.runbzr('merge ../b/')
-        assert os.path.exists('sub/a.txt.OTHER')
-        assert os.path.exists('sub/a.txt.BASE')
-
     def test_pull(self):
         """Pull changes from one branch to another."""
         os.mkdir('a')
@@ -728,12 +714,18 @@
             runbzr('remove d2/link1')
             assert self.capture('unknowns') == 'd2/link1\n'
             runbzr(['commit', '-m', '5: remove d2/link1'])
+            # try with the rm alias
+            runbzr('add d2/link1')
+            runbzr(['commit', '-m', '6: add d2/link1'])
+            runbzr('rm d2/link1')
+            assert self.capture('unknowns') == 'd2/link1\n'
+            runbzr(['commit', '-m', '7: remove d2/link1'])
     
             os.mkdir("d1")
             runbzr('add d1')
             runbzr('rename d2/link3 d1/link3new')
             assert self.capture('unknowns') == 'd2/link1\n'
-            runbzr(['commit', '-m', '6: remove d2/link1, move/rename link3'])
+            runbzr(['commit', '-m', '8: remove d2/link1, move/rename link3'])
             
             runbzr(['check'])
             
@@ -771,9 +763,9 @@
             assert listdir_sorted("d2")== [ "link3" ]
             chdir("..")
             
-            runbzr(['export', '-r', '6', 'exp6.tmp'])
+            runbzr(['export', '-r', '8', 'exp6.tmp'])
             chdir("exp6.tmp")
-            assert listdir_sorted(".") == [ "d1", "d2", "link2" ]
+            self.assertEqual(listdir_sorted("."), [ "d1", "d2", "link2"])
             assert listdir_sorted("d1") == [ "link3new" ]
             assert listdir_sorted("d2") == []
             assert os.readlink("d1/link3new") == "NOWHERE3"

=== modified file 'bzrlib/selftest/test_merge_core.py'
--- bzrlib/selftest/test_merge_core.py	2005-10-03 15:19:25 +0000
+++ bzrlib/selftest/test_merge_core.py	2005-10-10 03:42:29 +0000
@@ -11,7 +11,7 @@
 from bzrlib.osutils import file_kind, rename
 from bzrlib import changeset
 from bzrlib.merge_core import (ApplyMerge3, make_merge_changeset,
-                                BackupBeforeChange, PermissionsMerge)
+                               BackupBeforeChange, ExecFlagMerge)
 from bzrlib.changeset import Inventory, apply_changeset, invert_dict
 
 
@@ -223,11 +223,10 @@
             self.change_perms_tree(id, self.other, other)
 
         if base is not None or other is not None:
-            old_perms = os.stat(self.base.full_path(id)).st_mode &077
-            new_perms = os.stat(self.other.full_path(id)).st_mode &077
-            contents = changeset.ChangeUnixPermissions(old_perms, 
-                                                       new_perms)
-            self.cset.entries[id].metadata_change = contents
+            old_exec = bool(os.stat(self.base.full_path(id)).st_mode & 0111)
+            new_exec = bool(os.stat(self.other.full_path(id)).st_mode & 0111)
+            metadata = changeset.ChangeExecFlag(old_exec, new_exec)
+            self.cset.entries[id].metadata_change = metadata
 
     def change_name_tree(self, id, tree, name):
         new_path = tree.child_path(self.cset.entries[id].parent, name)
@@ -448,30 +447,21 @@
     def test_perms_merge(self):
         builder = MergeBuilder()
         builder.add_file("1", "0", "name1", "text1", 0755)
-        builder.change_perms("1", other=0655)
+        builder.change_perms("1", other=0644)
         builder.add_file("2", "0", "name2", "text2", 0755)
-        builder.change_perms("2", base=0655)
+        builder.change_perms("2", base=0644)
         builder.add_file("3", "0", "name3", "text3", 0755)
-        builder.change_perms("3", this=0655)
+        builder.change_perms("3", this=0644)
         cset = builder.merge_changeset(ApplyMerge3)
         assert(cset.entries["1"].metadata_change is not None)
-        assert(isinstance(cset.entries["1"].metadata_change,
-                          PermissionsMerge))
-        assert(isinstance(cset.entries["2"].metadata_change,
-                          PermissionsMerge))
+        assert(isinstance(cset.entries["1"].metadata_change, ExecFlagMerge))
+        assert(isinstance(cset.entries["2"].metadata_change, ExecFlagMerge))
         assert(cset.entries["3"].is_boring())
         builder.apply_changeset(cset)
-        assert(os.stat(builder.this.full_path("1")).st_mode &0777 == 0655)
+        assert(os.stat(builder.this.full_path("1")).st_mode &0777 == 0644)
         assert(os.stat(builder.this.full_path("2")).st_mode &0777 == 0755)
-        assert(os.stat(builder.this.full_path("3")).st_mode &0777 == 0655)
+        assert(os.stat(builder.this.full_path("3")).st_mode &0777 == 0644)
         builder.cleanup();
-        builder = MergeBuilder()
-        builder.add_file("1", "0", "name1", "text1", 0755)
-        builder.change_perms("1", other=0655, base=0555)
-        cset = builder.merge_changeset(ApplyMerge3)
-        self.assertRaises(changeset.MergePermissionConflict, 
-                     builder.apply_changeset, cset)
-        builder.cleanup()
 
 class FunctionalMergeTest(TestCaseInTempDir):
 

=== modified file 'bzrlib/selftest/test_upgrade.py'
--- bzrlib/selftest/test_upgrade.py	2005-10-06 11:26:51 +0000
+++ bzrlib/selftest/test_upgrade.py	2005-10-09 23:52:04 +0000
@@ -39,12 +39,12 @@
         self.failUnlessExists('.bzr/README')
 
     def test_upgrade_simple(self):
-        """Upgrade simple v0.0.4 format to v5"""
+        """Upgrade simple v0.0.4 format to v6"""
         eq = self.assertEquals
         build_tree_contents(_upgrade1_template)
         upgrade('.')
         b = Branch.open('.')
-        eq(b._branch_format, 5)
+        eq(b._branch_format, 6)
         rh = b.revision_history()
         eq(rh,
            ['mbp@sourcefrog.net-20051004035611-176b16534b086b3c',
@@ -77,8 +77,11 @@
     def test_upgrade_makes_dir_weaves(self):
         build_tree_contents(_upgrade_dir_template)
         upgrade('.')
+        # this is the path to the literal file. As format changes 
+        # occur it needs to be updated. FIXME: ask the store for the
+        # path.
         self.failUnlessExists(
-            '.bzr/weaves/dir-20051005095101-da1441ea3fa6917a.weave')
+            '.bzr/weaves/de/dir-20051005095101-da1441ea3fa6917a.weave')
 
 
 _upgrade1_template = \

=== modified file 'bzrlib/selftest/test_weave.py'
--- bzrlib/selftest/test_weave.py	2005-10-06 10:53:12 +0000
+++ bzrlib/selftest/test_weave.py	2005-10-08 00:03:38 +0000
@@ -18,14 +18,15 @@
 
 
 # TODO: tests regarding version names
-
-
+# TODO: rbc 20050108 test that join does not leave an inconsistent weave 
+#       if it fails.
 
 """test suite for weave algorithm"""
 
 from pprint import pformat
 
-from bzrlib.weave import Weave, WeaveFormatError, WeaveError
+import bzrlib.errors as errors
+from bzrlib.weave import Weave, WeaveFormatError, WeaveError, reweave
 from bzrlib.weavefile import write_weave, read_weave
 from bzrlib.selftest import TestCase
 from bzrlib.osutils import sha_string
@@ -873,7 +874,24 @@
         eq(sorted(wa.iter_names()), ['v1', 'v2', 'v3', 'x1',])
         eq(wa.get_text('x1'), 'line from x1\n')
 
-    def test_join_with_ghosts(self):
+    def test_reweave_with_empty(self):
+        wb = Weave()
+        wr = reweave(self.weave1, wb)
+        eq = self.assertEquals
+        eq(sorted(wr.iter_names()), ['v1', 'v2', 'v3'])
+        eq(wr.get_lines('v3'), ['hello\n', 'cruel\n', 'world\n'])
+        self.weave1.reweave(wb)
+        self.assertEquals(wr, self.weave1)
+
+    def test_join_with_ghosts_raises_parent_mismatch(self):
+        wa = self.weave1.copy()
+        wb = Weave()
+        wb.add('x1', [], ['line from x1\n'])
+        wb.add('v1', [], ['hello\n'])
+        wb.add('v2', ['v1', 'x1'], ['hello\n', 'world\n'])
+        self.assertRaises(errors.WeaveParentMismatch, wa.join, wb)
+
+    def test_reweave_with_ghosts(self):
         """Join that inserts parents of an existing revision.
 
         This can happen when merging from another branch who
@@ -881,16 +899,19 @@
         this test the second weave knows of an additional parent of 
         v2.  Any revisions which are in common still have to have the 
         same text."""
-        return ###############################
         wa = self.weave1.copy()
         wb = Weave()
         wb.add('x1', [], ['line from x1\n'])
         wb.add('v1', [], ['hello\n'])
         wb.add('v2', ['v1', 'x1'], ['hello\n', 'world\n'])
-        wa.join(wb)
+        wc = reweave(wa, wb)
         eq = self.assertEquals
-        eq(sorted(wa.iter_names()), ['v1', 'v2', 'v3', 'x1',])
-        eq(wa.get_text('x1'), 'line from x1\n')
+        eq(sorted(wc.iter_names()), ['v1', 'v2', 'v3', 'x1',])
+        eq(wc.get_text('x1'), 'line from x1\n')
+        eq(wc.get_lines('v2'), ['hello\n', 'world\n'])
+        eq(wc.parent_names('v2'), ['v1', 'x1'])
+        self.weave1.reweave(wb)
+        self.assertEquals(wc, self.weave1)
 
 
 if __name__ == '__main__':

=== modified file 'bzrlib/selftest/testbranch.py'
--- bzrlib/selftest/testbranch.py	2005-10-07 04:10:46 +0000
+++ bzrlib/selftest/testbranch.py	2005-10-09 22:43:56 +0000
@@ -19,11 +19,11 @@
 from bzrlib.clone import copy_branch
 from bzrlib.commit import commit
 import bzrlib.errors as errors
-from bzrlib.errors import NoSuchRevision, UnlistableBranch
+from bzrlib.errors import NoSuchRevision, UnlistableBranch, NotBranchError
 from bzrlib.selftest import TestCaseInTempDir
 from bzrlib.trace import mutter
 import bzrlib.transactions as transactions
-
+from bzrlib.selftest.HTTPTestUtil import TestCaseWithWebserver
 
 class TestBranch(TestCaseInTempDir):
 
@@ -140,8 +140,19 @@
                            'wibble@fofof--20050401--1928390812')
         # list should be cleared when we do a commit
         self.assertEquals(b.pending_merges(), [])
- 
-
+
+
+class TestRemote(TestCaseWithWebserver):
+
+    def test_open_containing(self):
+        self.assertRaises(NotBranchError, Branch.open_containing,
+                          self.get_remote_url(''))
+        self.assertRaises(NotBranchError, Branch.open_containing,
+                          self.get_remote_url('g/p/q'))
+        b = Branch.initialize('.')
+        Branch.open_containing(self.get_remote_url(''))
+        Branch.open_containing(self.get_remote_url('g/p/q'))
+        
 # TODO: rewrite this as a regular unittest, without relying on the displayed output        
 #         >>> from bzrlib.commit import commit
 #         >>> bzrlib.trace.silent = True

=== modified file 'bzrlib/selftest/testfetch.py'
--- bzrlib/selftest/testfetch.py	2005-10-06 06:31:33 +0000
+++ bzrlib/selftest/testfetch.py	2005-10-10 21:52:15 +0000
@@ -96,6 +96,8 @@
                       br_a2)
 
 
+    #TODO: test that fetch correctly does reweaving when needed. RBC 20051008
+
 class TestFetch(TestCaseInTempDir):
 
     def test_fetch(self):
@@ -133,6 +135,13 @@
         source = Branch.open(self.get_remote_url("source/"))
         source.weave_store.enable_cache = False
         self.assertEqual(greedy_fetch(target, source), (2, []))
-        weave_suffix = 'weaves/id.weave HTTP/1.1" 200 -'
+        # this is the path to the literal file. As format changes 
+        # occur it needs to be updated. FIXME: ask the store for the
+        # path.
+        weave_suffix = 'weaves/ce/id.weave HTTP/1.1" 200 -'
         self.assertEqual(1,
             len([log for log in self.weblogs if log.endswith(weave_suffix)]))
+        inventory_weave_suffix = 'inventory.weave HTTP/1.1" 200 -'
+        self.assertEqual(1,
+            len([log for log in self.weblogs if log.endswith(
+                inventory_weave_suffix)]))

=== modified file 'bzrlib/selftest/testinv.py'
--- bzrlib/selftest/testinv.py	2005-10-07 06:13:14 +0000
+++ bzrlib/selftest/testinv.py	2005-10-10 00:05:25 +0000
@@ -196,6 +196,8 @@
                                             "\n")
         
     def test_link_diff_deleted(self):
+        if not has_symlinks():
+            return
         output = StringIO()
         self.link_1.diff(internal_diff, 
                           "old_label", self.tree_1,
@@ -205,6 +207,8 @@
                          "=== target was 'target1'\n")
 
     def test_link_diff_added(self):
+        if not has_symlinks():
+            return
         output = StringIO()
         self.link_1.diff(internal_diff, 
                           "new_label", self.tree_1,
@@ -214,6 +218,8 @@
                          "=== target is 'target1'\n")
 
     def test_link_diff_changed(self):
+        if not has_symlinks():
+            return
         output = StringIO()
         self.link_1.diff(internal_diff, 
                           "/dev/null", self.tree_1, 

=== modified file 'bzrlib/selftest/testrevisionnamespaces.py'
--- bzrlib/selftest/testrevisionnamespaces.py	2005-09-29 20:34:25 +0000
+++ bzrlib/selftest/testrevisionnamespaces.py	2005-10-10 01:32:08 +0000
@@ -16,10 +16,14 @@
 
 import os
 import time
+
+from bzrlib.branch import Branch
 from bzrlib.selftest import TestCaseInTempDir
 from bzrlib.errors import NoCommonAncestor, NoCommits
+from bzrlib.errors import NoSuchRevision
 from bzrlib.clone import copy_branch
 from bzrlib.merge import merge
+from bzrlib.revisionspec import RevisionSpec
 
 class TestRevisionNamespaces(TestCaseInTempDir):
 
@@ -27,9 +31,6 @@
         """Test revision specifiers.
 
         These identify revisions by date, etc."""
-        from bzrlib.errors import NoSuchRevision
-        from bzrlib.branch import Branch
-        from bzrlib.revisionspec import RevisionSpec
 
         b = Branch.initialize('.')
 
@@ -78,3 +79,18 @@
                           'b@r-0-4')
         self.assertEquals(RevisionSpec('ancestor:.').in_history(b3).rev_id,
                           'b@r-0-4')
+
+    def test_branch_namespace(self):
+        """Ensure that the branch namespace pulls in the requisite content."""
+        self.build_tree(['branch1/', 'branch1/file', 'branch2/'])
+        branch = Branch.initialize('branch1')
+        branch.add(['file'])
+        branch.commit('add file')
+        copy_branch(branch, 'branch2')
+        print >> open('branch2/file', 'w'), 'new content'
+        branch2 = Branch.open('branch2')
+        branch2.commit('update file', rev_id='A')
+        spec = RevisionSpec('branch:./branch2/.bzr/../')
+        rev_info = spec.in_history(branch)
+        self.assertEqual(rev_info, (None, 'A'))
+

=== modified file 'bzrlib/store/__init__.py'
--- bzrlib/store/__init__.py	2005-10-04 04:37:29 +0000
+++ bzrlib/store/__init__.py	2005-10-10 03:24:15 +0000
@@ -25,7 +25,10 @@
 """
 
 from cStringIO import StringIO
+from stat import ST_MODE, S_ISDIR
+from zlib import adler32
 
+import bzrlib.errors as errors
 from bzrlib.errors import BzrError, UnlistableStore, TransportNotPossible
 from bzrlib.trace import mutter
 import bzrlib.transport
@@ -191,6 +194,14 @@
 
     _max_buffered_requests = 10
 
+    def __getitem__(self, fileid):
+        """Returns a file reading from a particular entry."""
+        fn = self._relpath(fileid)
+        try:
+            return self._transport.get(fn)
+        except errors.NoSuchFile:
+            raise KeyError(fileid)
+
     def __init__(self, transport):
         assert isinstance(transport, bzrlib.transport.Transport)
         super(TransportStore, self).__init__()
@@ -204,6 +215,19 @@
 
     __str__ = __repr__
 
+    def _iter_relpaths(self):
+        """Iter the relative paths of files in the transports sub-tree."""
+        transport = self._transport
+        queue = list(transport.list_dir('.'))
+        while queue:
+            relpath = queue.pop(0)
+            st = transport.stat(relpath)
+            if S_ISDIR(st[ST_MODE]):
+                for i, basename in enumerate(transport.list_dir(relpath)):
+                    queue.insert(i, relpath+'/'+basename)
+            else:
+                yield relpath, st
+
     def listable(self):
         """Return True if this store is able to be listed."""
         return self._transport.listable()
@@ -307,3 +331,6 @@
     ids = [f for f in store_from]
     store_to.copy_multi(store_from, ids)
 
+def hash_prefix(file_id):
+    return "%02x/" % (adler32(file_id) & 0xff)
+

=== modified file 'bzrlib/store/compressed_text.py'
--- bzrlib/store/compressed_text.py	2005-10-02 21:51:29 +0000
+++ bzrlib/store/compressed_text.py	2005-10-10 03:24:15 +0000
@@ -24,8 +24,9 @@
 import os, tempfile, gzip
 
 import bzrlib.store
+from bzrlib.store import hash_prefix
 from bzrlib.trace import mutter
-from bzrlib.errors import BzrError
+from bzrlib.errors import BzrError, FileExists
 
 from StringIO import StringIO
 from stat import ST_SIZE
@@ -56,8 +57,9 @@
     'goodbye'
     """
 
-    def __init__(self, transport):
+    def __init__(self, transport, prefixed=False):
         super(CompressedTextStore, self).__init__(transport)
+        self._prefixed = prefixed
 
     def _check_fileid(self, fileid):
         if '\\' in fileid or '/' in fileid:
@@ -65,7 +67,10 @@
 
     def _relpath(self, fileid):
         self._check_fileid(fileid)
-        return fileid + '.gz'
+        if self._prefixed:
+            return hash_prefix(fileid) + fileid + ".gz"
+        else:
+            return fileid + ".gz"
 
     def add(self, f, fileid):
         """Add contents of a file into the store.
@@ -88,6 +93,11 @@
         if self._transport.has(fn):
             raise BzrError("store %r already contains id %r" % (self._transport.base, fileid))
 
+        if self._prefixed:
+            try:
+                self._transport.mkdir(hash_prefix(fileid))
+            except FileExists:
+                pass
 
         sio = StringIO()
         gf = gzip.GzipFile(mode='wb', fileobj=sio)
@@ -197,26 +207,18 @@
             count += 1
 
     def __iter__(self):
-        # TODO: case-insensitive?
-        for f in self._transport.list_dir('.'):
-            if f[-3:] == '.gz':
-                yield f[:-3]
+        for relpath, st in self._iter_relpaths():
+            if relpath.endswith(".gz"):
+                yield os.path.basename(relpath)[:-3]
             else:
-                yield f
+                yield os.path.basename(relpath)
 
     def __len__(self):
-        return len([f for f in self._transport.list_dir('.')])
-
+        return len(list(self._iter_relpath()))
 
     def __getitem__(self, fileid):
         """Returns a file reading from a particular entry."""
-        fn = self._relpath(fileid)
-        # This will throw if the file doesn't exist.
-        try:
-            f = self._transport.get(fn)
-        except:
-            raise KeyError('This store (%s) does not contain %s' % (self, fileid))
-
+        f = super(CompressedTextStore, self).__getitem__(fileid)
         # gzip.GzipFile.read() requires a tell() function
         # but some transports return objects that cannot seek
         # so buffer them in a StringIO instead
@@ -226,7 +228,6 @@
             from cStringIO import StringIO
             sio = StringIO(f.read())
             return gzip.GzipFile(mode='rb', fileobj=sio)
-            
 
     def total_size(self):
         """Return (count, bytes)
@@ -235,8 +236,7 @@
         the content."""
         total = 0
         count = 0
-        relpaths = [self._relpath(fid) for fid in self]
-        for st in self._transport.stat_multi(relpaths):
+        for relpath, st in self._iter_relpaths():
             count += 1
             total += st[ST_SIZE]
                 

=== modified file 'bzrlib/store/text.py'
--- bzrlib/store/text.py	2005-10-02 21:51:29 +0000
+++ bzrlib/store/text.py	2005-10-10 03:24:15 +0000
@@ -24,8 +24,9 @@
 import os, tempfile
 
 import bzrlib.store
+from bzrlib.store import hash_prefix
 from bzrlib.trace import mutter
-from bzrlib.errors import BzrError
+from bzrlib.errors import BzrError, FileExists
 
 from cStringIO import StringIO
 from stat import ST_SIZE
@@ -41,8 +42,9 @@
     Files are stored uncompressed, with no delta compression.
     """
 
-    def __init__(self, transport):
+    def __init__(self, transport, prefixed=False):
         super(TextStore, self).__init__(transport)
+        self._prefixed = prefixed
 
     def _check_fileid(self, fileid):
         if not isinstance(fileid, basestring):
@@ -52,7 +54,10 @@
 
     def _relpath(self, fileid):
         self._check_fileid(fileid)
-        return fileid
+        if self._prefixed:
+            return hash_prefix(fileid) + fileid
+        else:
+            return fileid
 
     def add(self, f, fileid):
         """Add contents of a file into the store.
@@ -65,38 +70,14 @@
         if self._transport.has(fn):
             raise BzrError("store %r already contains id %r" % (self._transport.base, fileid))
 
+        if self._prefixed:
+            try:
+                self._transport.mkdir(hash_prefix(fileid))
+            except FileExists:
+                pass
+
         self._transport.put(fn, f)
 
-    def _do_copy(self, other, to_copy, pb, permit_failure=False):
-        if isinstance(other, TextStore):
-            return self._copy_multi_text(other, to_copy, pb,
-                    permit_failure=permit_failure)
-        return super(TextStore, self)._do_copy(other, to_copy,
-                pb, permit_failure=permit_failure)
-
-    def _copy_multi_text(self, other, to_copy, pb,
-            permit_failure=False):
-        # Because of _transport, we can no longer assume
-        # that they are on the same filesystem, we can, however
-        # assume that we only need to copy the exact bytes,
-        # we don't need to process the files.
-
-        failed = set()
-        if permit_failure:
-            new_to_copy = set()
-            for fileid, has in zip(to_copy, other.has(to_copy)):
-                if has:
-                    new_to_copy.add(fileid)
-                else:
-                    failed.add(fileid)
-            to_copy = new_to_copy
-            #mutter('_copy_multi_text copying %s, failed %s' % (to_copy, failed))
-
-        paths = [self._relpath(fileid) for fileid in to_copy]
-        count = other._transport.copy_to(paths, self._transport, pb=pb)
-        assert count == len(to_copy)
-        return count, failed
-
     def __contains__(self, fileid):
         """"""
         fn = self._relpath(fileid)
@@ -157,25 +138,11 @@
             count += 1
 
     def __iter__(self):
-        # TODO: case-insensitive?
-        for f in self._transport.list_dir('.'):
-            yield f
+        for relpath, st in self._iter_relpaths():
+            yield os.path.basename(relpath)
 
     def __len__(self):
-        return len([f for f in self._transport.list_dir('.')])
-
-
-    def __getitem__(self, fileid):
-        """Returns a file reading from a particular entry."""
-        fn = self._relpath(fileid)
-        # This will throw if the file doesn't exist.
-        try:
-            f = self._transport.get(fn)
-        except:
-            raise KeyError('This store (%s) does not contain %s' % (self, fileid))
-
-        return f
-            
+        return len(list(self._iter_relpath()))
 
     def total_size(self):
         """Return (count, bytes)
@@ -184,8 +151,7 @@
         the content."""
         total = 0
         count = 0
-        relpaths = [self._relpath(fid) for fid in self]
-        for st in self._transport.stat_multi(relpaths):
+        for relpath, st in self._iter_relpaths():
             count += 1
             total += st[ST_SIZE]
                 

=== modified file 'bzrlib/store/weave.py'
--- bzrlib/store/weave.py	2005-10-07 06:13:14 +0000
+++ bzrlib/store/weave.py	2005-10-10 03:24:15 +0000
@@ -26,32 +26,35 @@
 
 from bzrlib.weavefile import read_weave, write_weave_v5
 from bzrlib.weave import Weave
-from bzrlib.store import Store
+from bzrlib.store import TransportStore, hash_prefix
 from bzrlib.atomicfile import AtomicFile
-from bzrlib.errors import NoSuchFile
+from bzrlib.errors import NoSuchFile, FileExists
 from bzrlib.trace import mutter
 
 
-class WeaveStore(Store):
+class WeaveStore(TransportStore):
     """Collection of several weave files in a directory.
 
     This has some shortcuts for reading and writing them.
     """
     FILE_SUFFIX = '.weave'
 
-    def __init__(self, transport):
+    def __init__(self, transport, prefixed=False):
         self._transport = transport
+        self._prefixed = prefixed
 
     def filename(self, file_id):
         """Return the path relative to the transport root."""
-        return file_id + WeaveStore.FILE_SUFFIX
+        if self._prefixed:
+            return hash_prefix(file_id) + file_id + WeaveStore.FILE_SUFFIX
+        else:
+            return file_id + WeaveStore.FILE_SUFFIX
 
     def __iter__(self):
         l = len(WeaveStore.FILE_SUFFIX)
-        for f in self._transport.list_dir('.'):
-            if f.endswith(WeaveStore.FILE_SUFFIX):
-                f = f[:-l]
-                yield f
+        for relpath, st in self._iter_relpaths():
+            if relpath.endswith(WeaveStore.FILE_SUFFIX):
+                yield os.path.basename(relpath[:-l])
 
     def __contains__(self, fileid):
         """"""
@@ -61,6 +64,11 @@
         return self._transport.get(self.filename(file_id))
 
     def _put(self, file_id, f):
+        if self._prefixed:
+            try:
+                self._transport.mkdir(hash_prefix(file_id))
+            except FileExists:
+                pass
         return self._transport.put(self.filename(file_id), f)
 
     def get_weave(self, file_id, transaction):
@@ -73,7 +81,6 @@
         transaction.register_clean(w)
         return w
 
-
     def get_lines(self, file_id, rev_id, transaction):
         """Return text from a particular version of a weave.
 
@@ -99,10 +106,8 @@
         sio = StringIO()
         write_weave_v5(weave, sio)
         sio.seek(0)
-
         self._put(file_id, sio)
 
-
     def add_text(self, file_id, rev_id, new_lines, parents, transaction):
         w = self.get_weave_or_empty(file_id, transaction)
         parent_idxs = map(w.lookup, parents)

=== modified file 'bzrlib/trace.py'
--- bzrlib/trace.py	2005-10-06 07:28:41 +0000
+++ bzrlib/trace.py	2005-10-10 05:12:40 +0000
@@ -161,13 +161,15 @@
     The exception string representation is used as the error
     summary, unless msg is given.
     """
+    command = ' '.join(repr(arg) for arg in sys.argv)
+    prefix = "command: %s\npwd: %s\n" % (command, os.getcwd())
     if msg == None:
         ei = sys.exc_info()
         msg = str(ei[1])
     if msg and (msg[-1] == '\n'):
         msg = msg[:-1]
     ## msg = "(%s) %s" % (str(type(ei[1])), msg)
-    _bzr_logger.exception(msg)
+    _bzr_logger.exception(prefix + msg)
 
 
 

=== modified file 'bzrlib/transport/http.py'
--- bzrlib/transport/http.py	2005-10-04 04:37:29 +0000
+++ bzrlib/transport/http.py	2005-10-10 03:24:15 +0000
@@ -89,7 +89,8 @@
                     # In most filesystems, a request for the parent
                     # of root, just returns root.
                     continue
-                basepath.pop()
+                if len(basepath) > 0:
+                    basepath.pop()
             elif p == '.':
                 continue # No-op
             else:
@@ -144,9 +145,9 @@
         try:
             return get_url(self.abspath(relpath))
         except (BzrError, urllib2.URLError, IOError), e:
-            raise NoSuchFile(orig_error=e)
-        except Exception,e:
-            raise HttpTransportError(orig_error=e)
+            raise NoSuchFile(msg = "Error retrieving %s" 
+                             % self.abspath(relpath),
+                             orig_error=e)
 
     def get_partial(self, relpath, start, length=None):
         """Get just part of a file.

=== modified file 'bzrlib/upgrade.py'
--- bzrlib/upgrade.py	2005-10-06 06:31:33 +0000
+++ bzrlib/upgrade.py	2005-10-09 23:42:12 +0000
@@ -85,7 +85,8 @@
 import logging
 import shutil
 
-from bzrlib.branch import Branch, find_branch, BZR_BRANCH_FORMAT_5
+from bzrlib.branch import Branch, find_branch
+from bzrlib.branch import BZR_BRANCH_FORMAT_5, BZR_BRANCH_FORMAT_6
 from bzrlib.revfile import Revfile
 from bzrlib.weave import Weave
 from bzrlib.weavefile import read_weave, write_weave
@@ -112,9 +113,38 @@
             return
         note('starting upgrade of %s', os.path.abspath(self.base))
         self._backup_control_dir()
-        note('starting upgrade')
+        self.pb = ui_factory.progress_bar()
+        if self.old_format == 4:
+            note('starting upgrade from format 4 to 5')
+            self._convert_to_weaves()
+            self._open_branch()
+        if self.old_format == 5:
+            note('starting upgrade from format 5 to 6')
+            self._convert_to_prefixed()
+            self._open_branch()
+        note("finished")
+
+
+    def _convert_to_prefixed(self):
+        from bzrlib.store import hash_prefix
+        for store_name in ["weaves", "revision-store"]:
+            note("adding prefixes to %s" % store_name) 
+            store_dir = os.path.join(self.base, ".bzr", store_name)
+            for filename in os.listdir(store_dir):
+                if filename.endswith(".weave") or filename.endswith(".gz"):
+                    file_id = os.path.splitext(filename)[0]
+                else:
+                    file_id = filename
+                prefix_dir = os.path.join(store_dir, hash_prefix(file_id))
+                if not os.path.isdir(prefix_dir):
+                    os.mkdir(prefix_dir)
+                os.rename(os.path.join(store_dir, filename),
+                          os.path.join(prefix_dir, filename))
+        self._set_new_format(BZR_BRANCH_FORMAT_6)
+
+
+    def _convert_to_weaves(self):
         note('note: upgrade may be faster if all store files are ungzipped first')
-        self.pb = ui_factory.progress_bar()
         if not os.path.isdir(self.base + '/.bzr/weaves'):
             os.mkdir(self.base + '/.bzr/weaves')
         self.inv_weave = Weave('inventory')
@@ -144,23 +174,24 @@
         note('  %6d texts' % self.text_count)
         self._write_all_weaves()
         self._write_all_revs()
-        self._set_new_format()
         self._cleanup_spare_files()
+        self._set_new_format(BZR_BRANCH_FORMAT_5)
 
 
     def _open_branch(self):
         self.branch = Branch.open_downlevel(self.base)
-        if self.branch._branch_format == 5:
-            note('this branch is already in the most current format')
+        self.old_format = self.branch._branch_format
+        if self.old_format == 6:
+            note('this branch is in the most current format')
             return False
-        if self.branch._branch_format != 4:
+        if self.old_format not in (4, 5):
             raise BzrError("cannot upgrade from branch format %r" %
                            self.branch._branch_format)
         return True
 
 
-    def _set_new_format(self):
-        self.branch.put_controlfile('branch-format', BZR_BRANCH_FORMAT_5)
+    def _set_new_format(self, format):
+        self.branch.put_controlfile('branch-format', format)
 
 
     def _cleanup_spare_files(self):

=== modified file 'bzrlib/weave.py'
--- bzrlib/weave.py	2005-10-06 08:11:46 +0000
+++ bzrlib/weave.py	2005-10-08 00:03:38 +0000
@@ -74,7 +74,7 @@
 import sha
 from difflib import SequenceMatcher
 
-
+from bzrlib.trace import mutter
 
 
 class WeaveError(Exception):
@@ -83,6 +83,10 @@
 
 class WeaveFormatError(WeaveError):
     """Weave invariant violated"""
+
+
+class WeaveParentMismatch(WeaveError):
+    """Parents are mismatched between two revisions."""
     
 
 class Weave(object):
@@ -398,6 +402,11 @@
         return self._parents[version]
 
 
+    def parent_names(self, version):
+        """Return version names for parents of a version."""
+        return map(self.idx_to_name, self._parents[self.lookup(version)])
+
+
     def minimal_parents(self, version):
         """Find the minimal set of parents for the version."""
         included = self._parents[version]
@@ -774,17 +783,20 @@
         retrieved from self after this call.
 
         It is illegal for the two weaves to contain different values 
-        for any version.
+        or different parents for any version.  See also reweave().
         """
         if other.numversions() == 0:
             return          # nothing to update, easy
+        # two loops so that we do not change ourselves before verifying it
+        # will be ok
         # work through in index order to make sure we get all dependencies
         for other_idx, name in enumerate(other._names):
+            if self._check_version_consistent(other, other_idx, name):
+                continue
+        for other_idx, name in enumerate(other._names):
             # TODO: If all the parents of the other version are already 
             # present then we can avoid some work by just taking the delta
             # and adjusting the offsets.
-            if self._check_version_consistent(other, other_idx, name):
-                continue
             new_parents = self._imported_parents(other, other_idx)
             lines = other.get_lines(other_idx)
             sha1 = other._sha1s[other_idx]
@@ -828,14 +840,98 @@
             n1.sort()
             n2.sort()
             if n1 != n2:
-                raise WeaveError("inconsistent parents for version {%s}: "
-                                 "%s vs %s"
-                                 % (name, n1, n2))
+                raise WeaveParentMismatch("inconsistent parents "
+                    "for version {%s}: %s vs %s" % (name, n1, n2))
             else:
                 return True         # ok!
         else:
             return False
 
+    def reweave(self, other):
+        """Reweave self with other."""
+        new_weave = reweave(self, other)
+        for attr in self.__slots__:
+            setattr(self, attr, getattr(new_weave, attr))
+
+
+def reweave(wa, wb):
+    """Combine two weaves and return the result.
+
+    This works even if a revision R has different parents in 
+    wa and wb.  In the resulting weave all the parents are given.
+
+    This is done by just building up a new weave, maintaining ordering 
+    of the versions in the two inputs.  More efficient approaches
+    might be possible but it should only be necessary to do 
+    this operation rarely, when a new previously ghost version is 
+    inserted.
+    """
+    wr = Weave()
+    ia = ib = 0
+    queue_a = range(wa.numversions())
+    queue_b = range(wb.numversions())
+    # first determine combined parents of all versions
+    # map from version name -> all parent names
+    combined_parents = _reweave_parent_graphs(wa, wb)
+    mutter("combined parents: %r", combined_parents)
+    order = _make_reweave_order(wa._names, wb._names, combined_parents)
+    mutter("order to reweave: %r", order)
+    for name in order:
+        if name in wa._name_map:
+            lines = wa.get_lines(name)
+            if name in wb._name_map:
+                assert lines == wb.get_lines(name)
+        else:
+            lines = wb.get_lines(name)
+        wr.add(name, combined_parents[name], lines)
+    return wr
+
+
+def _reweave_parent_graphs(wa, wb):
+    """Return combined parent ancestry for two weaves.
+    
+    Returned as a list of (version_name, set(parent_names))"""
+    combined = {}
+    for weave in [wa, wb]:
+        for idx, name in enumerate(weave._names):
+            p = combined.setdefault(name, set())
+            p.update(map(weave.idx_to_name, weave._parents[idx]))
+    return combined
+
+
+def _make_reweave_order(wa_order, wb_order, combined_parents):
+    """Return an order to reweave versions respecting parents."""
+    done = set()
+    result = []
+    ia = ib = 0
+    next_a = next_b = None
+    len_a = len(wa_order)
+    len_b = len(wb_order)
+    while ia < len(wa_order) or ib < len(wb_order):
+        if ia < len_a:
+            next_a = wa_order[ia]
+            if next_a in done:
+                ia += 1
+                continue
+            if combined_parents[next_a].issubset(done):
+                ia += 1
+                result.append(next_a)
+                done.add(next_a)
+                continue
+        if ib < len_b:
+            next_b = wb_order[ib]
+            if next_b in done:
+                ib += 1
+                continue
+            elif combined_parents[next_b].issubset(done):
+                ib += 1
+                result.append(next_b)
+                done.add(next_b)
+                continue
+        raise WeaveError("don't know how to reweave at {%s} and {%s}"
+                         % (next_a, next_b))
+    return result
+
 
 def weave_toc(w):
     """Show the weave's table-of-contents"""

=== modified file 'bzrlib/workingtree.py'
--- bzrlib/workingtree.py	2005-10-06 04:29:10 +0000
+++ bzrlib/workingtree.py	2005-10-10 06:57:08 +0000
@@ -185,7 +185,7 @@
             return bool(stat.S_ISREG(mode) and stat.S_IEXEC&mode)
 
     def get_symlink_target(self, file_id):
-        return os.readlink(self.id2path(file_id))
+        return os.readlink(self.id2abspath(file_id))
 
     def file_class(self, filename):
         if self.path2id(filename):

=== modified file 'setup.py' (properties changed: -x to +x)
